r"""Tests for 3D Delaunay Triangulation.

.. todo:
   * answer test circumcenter
   * answer test ability to flip
   * clarify equality between facets defined using different cells

"""
import numpy as np
import os
from cgal4py.delaunay import Delaunay3
from test_cgal4py import MyTestCase, make_points


pts, left_edge, right_edge = make_points(0, 3)
pts_dup = np.concatenate([pts, np.reshape(pts[0, :], (1, pts.shape[1]))])
nverts_fin = pts.shape[0]
nverts_inf = 1
nverts = nverts_fin + nverts_inf
nedges_fin = 26
nedges_inf = 8
nedges = nedges_fin + nedges_inf
ncells_fin = 12
ncells_inf = 12
ncells = ncells_fin + ncells_inf
nfacets_fin = 30
nfacets_inf = 18
nfacets = nfacets_fin + nfacets_inf


def test_create():
    T = Delaunay3()
    del T


def test_insert():
    # without duplicates
    T = Delaunay3()
    T.insert(pts)
    assert(T.is_valid())
    # with duplicates
    T = Delaunay3()
    T.insert(pts_dup)
    assert(T.is_valid())


def test_equal():
    T1 = Delaunay3()
    T1.insert(pts)
    T2 = Delaunay3()
    T2.insert(pts)
    assert(T1.is_equivalent(T2))


def test_num_verts():
    # without duplicates
    T = Delaunay3()
    T.insert(pts)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)
    # with duplicates
    T = Delaunay3()
    T.insert(pts_dup)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)


def test_num_edges():
    # without duplicates
    T = Delaunay3()
    T.insert(pts)
    print(T.num_finite_edges, T.num_infinite_edges, T.num_edges)
    assert(T.num_finite_edges == nedges_fin)
    assert(T.num_infinite_edges == nedges_inf)
    assert(T.num_edges == nedges)
    # with duplicates
    T = Delaunay3()
    T.insert(pts_dup)
    print(T.num_finite_edges, T.num_infinite_edges, T.num_edges)
    assert(T.num_finite_edges == nedges_fin)
    assert(T.num_infinite_edges == nedges_inf)
    assert(T.num_edges == nedges)


def test_num_facets():
    # without duplicates
    T = Delaunay3()
    T.insert(pts)
    print(T.num_finite_facets, T.num_infinite_facets, T.num_facets)
    assert(T.num_finite_facets == nfacets_fin)
    assert(T.num_infinite_facets == nfacets_inf)
    assert(T.num_facets == nfacets)
    # with duplicates
    T = Delaunay3()
    T.insert(pts_dup)
    print(T.num_finite_facets, T.num_infinite_facets, T.num_facets)
    assert(T.num_finite_facets == nfacets_fin)
    assert(T.num_infinite_facets == nfacets_inf)
    assert(T.num_facets == nfacets)


def test_num_cells():
    # without duplicates
    T = Delaunay3()
    T.insert(pts)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)
    # with duplicates
    T = Delaunay3()
    T.insert(pts_dup)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)


def test_all_verts():
    T = Delaunay3()
    T.insert(pts)
    count_fin = count_inf = 0
    for v in T.all_verts:
        if v.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_finite_verts)
    assert(count_inf == T.num_infinite_verts)
    assert(count == T.num_verts)


def test_finite_verts():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.finite_verts:
        assert((not v.is_infinite()))
        count += 1
    assert(count == T.num_finite_verts)


def test_all_edges():
    T = Delaunay3()
    T.insert(pts)
    count_fin = count_inf = 0
    for e in T.all_edges:
        if e.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    print(count_fin, count_inf)
    assert(count_fin == T.num_finite_edges)
    assert(count_inf == T.num_infinite_edges)
    assert(count == T.num_edges)


def test_finite_edges():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for e in T.finite_edges:
        assert((not e.is_infinite()))
        count += 1
    print(count)
    assert(count == T.num_finite_edges)


def test_all_cells():
    T = Delaunay3()
    T.insert(pts)
    count_fin = count_inf = 0
    for c in T.all_cells:
        if c.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_finite_cells)
    assert(count_inf == T.num_infinite_cells)
    assert(count == T.num_cells)


def test_finite_cells():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for c in T.finite_cells:
        assert((not c.is_infinite()))
        count += 1
    assert(count == T.num_finite_cells)


def test_get_vertex():
    T = Delaunay3()
    T.insert(pts)
    for i in range(nverts_fin):
        v = T.get_vertex(i)
        assert(np.allclose(v.point, pts[i, :]))


def test_locate():
    T = Delaunay3()
    T.insert(pts)
    for c in T.finite_cells:
        p = c.center
        print(c, p, T.locate(p))
        assert(c == T.locate(p))
        assert(c == T.locate(p, c))
        assert(c.vertex(0) == T.locate(c.vertex(0).point))
        assert(c.vertex(0) == T.locate(c.vertex(0).point, c))
        assert(c.facet(0).is_equivalent(T.locate(c.facet(0).center)))
        assert(c.facet(0).is_equivalent(T.locate(c.facet(0).center, c)))
        # assert(c.facet(0) == T.locate(c.facet(0).center))
        # assert(c.facet(0) == T.locate(c.facet(0).center, c))
        assert(c.facet(0).edge(0).is_equivalent(
            T.locate(c.facet(0).edge(0).center)))
        assert(c.facet(0).edge(0).is_equivalent(
            T.locate(c.facet(0).edge(0).center, c)))
        # assert(c.facet(0).edge(0) == T.locate(c.facet(0).edge(0).center))
        # assert(c.facet(0).edge(0) == T.locate(c.facet(0).edge(0).center, c))
        break


def test_remove():
    T = Delaunay3()
    T.insert(pts)
    v = T.get_vertex(0)
    T.remove(v)
    assert(T.num_verts == (nverts-1))


def test_clear():
    T = Delaunay3()
    T.insert(pts)
    T.clear()
    print(T.num_finite_verts, T.num_cells)
    assert(T.num_finite_verts == 0)
    assert(T.num_cells == 0)


def test_is_edge():
    T = Delaunay3()
    T.insert(pts)
    assert(T.is_edge(T.get_vertex(0), T.get_vertex(1)))
    assert(not T.is_edge(T.get_vertex(1), T.get_vertex(nverts_fin-1)))


def test_is_facet():
    T = Delaunay3()
    T.insert(pts)
    assert(T.is_facet(T.get_vertex(0), T.get_vertex(1), T.get_vertex(3)))
    assert(not T.is_facet(T.get_vertex(0), T.get_vertex(1),
                          T.get_vertex(nverts_fin-1)))


def test_is_cell():
    T = Delaunay3()
    T.insert(pts)
    assert(T.is_cell(T.get_vertex(0), T.get_vertex(1),
                     T.get_vertex(3), T.get_vertex(5)))
    assert(not T.is_cell(T.get_vertex(0), T.get_vertex(1),
                         T.get_vertex(3), T.get_vertex(nverts_fin-1)))


def test_vert():
    T = Delaunay3()
    T.insert(pts)
    vold = None
    for v in T.all_verts:
        idx = v.index
        pnt = v.point
        vol = v.dual_volume
        print(v, idx, pnt, vol)
        assert(v == v)
        if vold is not None:
            assert(v != vold)
        if v.is_infinite():
            assert(idx == np.iinfo(np.uint32).max)
            assert(np.isinf(pnt).all())
            assert(np.isclose(vol, -1.0))
        else:
            assert(np.allclose(pnt, pts[idx, :]))
            if idx == 0:
                assert(np.isclose(vol, 4.5))
            else:
                assert(np.isclose(vol, -1.0))
            c = v.cell
            v.set_cell(c)
            v.set_point(pnt)
        vold = v


def test_edge():
    T = Delaunay3()
    T.insert(pts)
    eold = None
    for e in T.all_edges:
        v1 = e.vertex(0)
        v2 = e.vertex(1)
        assert(v1 == e.vertex1)
        assert(v2 == e.vertex2)
        c = e.cell
        i1 = e.ind1
        i2 = e.ind2
        elen = e.length
        inf = e.is_infinite()
        gab = e.is_Gabriel()
        print(e, v1.index, v2.index, elen, inf, gab)
        assert(e == e)
        assert(e.is_equivalent(e))
        if eold is not None:
            assert(e != eold)
        p1 = e.center
        p2 = v1.point
        print(e.side(p1), p1)
        print(e.side(p2), p2)
        if e.is_infinite():
            assert(np.isclose(elen, -1.0))
            assert(np.isinf(e.center).all())
            assert(e.side(p1) == -1)
            assert(e.side(p2) == -1)
        else:
            l = np.sqrt(np.sum((pts[v1.index, :]-pts[v2.index, :])**2.0))
            assert(np.isclose(elen, l))
            # p3 = e.center + 10*elen
            # print(e.side(p3), p3)
            # assert(e.side(p1) == -1) # virtually impossible
            # assert(e.side(p2) == 0)
            # assert(e.side(p3) == 1)
        eold = e
        del(c, i1, i2)


def test_facet():
    T = Delaunay3()
    T.insert(pts)
    fold = None
    for f in T.all_facets:
        v1 = f.vertex(0)
        v2 = f.vertex(1)
        v3 = f.vertex(2)
        e1 = f.edge(0)
        e2 = f.edge(1)
        e3 = f.edge(2)
        c = f.cell
        i = f.ind
        inf = f.is_infinite()
        gab = f.is_Gabriel()
        print(f, v1.index, v2.index, v3.index, i, inf, gab, f.center)
        assert(f == f)
        assert(f.is_equivalent(f))
        if fold is not None:
            assert(f != fold)
        del(e1, e2, e3, c)

        p1 = f.center
        p2 = v1.point
        print(f.side(p1), p1)
        print(f.side(p2), p2)
        if f.is_infinite():
            assert(np.isinf(f.center).all())
            assert(f.side(p1) == -1)
            assert(f.side(p2) == -1)
        # else:
        #     p3 = 2*v1.point - f.center + np.arange(3)
        #     print(f.side(p3), p3)
        #     assert(f.side(p1) == -1)
        #     assert(f.side(p2) == 0)
        #     assert(f.side(p3) == 1)

        # # This segfaults inside CGAL function call
        # print(f.side_of_circle((v1.point+v2.point+v3.point)/3),
        #       (v1.point+v2.point+v3.point)/3)
        # print(f.side_of_circle(v1.point), v1.point)
        # print(f.side_of_circle((5*v1.point-v2.point-v3.point)/3),
        #       (5*v1.point-v2.point-v3.point)/3)
        # if f.is_infinite():
        #     assert(f.side_of_circle((v1.point+v2.point+v3.point)/3) == -1)
        #     assert(f.side_of_circle(v1.point) == -1)
        #     assert(f.side_of_circle((5*v1.point-v2.point-v3.point)/3) == -1)
        # else:
        #     # This segfaults...
        #     assert(f.side_of_circle((v1.point+v2.point+v3.point)/3) == -1)
        #     assert(f.side_of_circle(v1.point) == 0)
        #     assert(f.side_of_circle((5*v1.point-v2.point-v3.point)/3) == 1)

        fold = f


def test_cell():
    T = Delaunay3()
    T.insert(pts)
    cold = None
    for c in T.all_cells:
        print(c, c.circumcenter, c.center)
        assert(c == c)
        if cold is not None:
            assert(c != cold)

        f1 = c.facet(0)
        f2 = c.facet(1)
        f3 = c.facet(2)
        f4 = c.facet(3)
        del(f1, f2, f3, f4)

        v1 = c.vertex(0)
        v2 = c.vertex(1)
        v3 = c.vertex(2)
        v4 = c.vertex(3)
        assert(c.has_vertex(v1))
        assert(c.has_vertex(v1, return_index=True) == 0)
        assert(c.ind_vertex(v1) == 0)

        c.reset_vertices()
        c.set_vertex(0, v1)
        c.set_vertices(v4, v3, v2, v1)

        n1 = c.neighbor(0)
        n2 = c.neighbor(1)
        n3 = c.neighbor(2)
        n4 = c.neighbor(3)
        assert(c.has_neighbor(n1))
        assert(c.has_neighbor(n1, return_index=True) == 0)
        assert(c.ind_neighbor(n1) == 0)

        c.reset_neighbors()
        c.set_neighbor(0, n1)
        c.set_neighbors(n4, n3, n2, n1)

        p1 = c.center
        p2 = v1.point
        print(c.side(p1), p1)
        print(c.side(p2), p2)
        print(c.side_of_sphere(c.circumcenter), c.circumcenter)
        print(c.side_of_sphere(v1.point), v1.point)
        if c.is_infinite():
            assert(np.isinf(c.center).all())
            assert(c.side(p1) == -1)
            assert(c.side(p2) == -1)
            assert(np.isinf(c.circumcenter).all())
            assert(c.side_of_sphere(c.circumcenter) == -1)
            assert(c.side_of_sphere(v1.point) == -1)
        else:
            p3 = 2*v1.point - c.center
            print(c.side(p3), p3)
            print(c.side_of_sphere(2*v1.point - c.circumcenter),
                  2*v1.point - c.circumcenter)
            assert(c.side(p1) == -1)
            assert(c.side(p2) == 0)
            assert(c.side(p3) == 1)
            assert(c.side_of_sphere(c.circumcenter) == -1)
            assert(c.side_of_sphere(v1.point) == 0)
            assert(c.side_of_sphere(2*v1.point - c.circumcenter) == 1)
        cold = c


def test_move():
    T = Delaunay3()
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(3, 'float64')
    v = T.move(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(T.num_verts == (nverts-1))


def test_move_if_no_collision():
    T = Delaunay3()
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(3, 'float64')
    v = T.move_if_no_collision(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move_if_no_collision(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v1.point, pts[1, :]))
    assert(T.num_verts == nverts)


def test_flip():
    T = Delaunay3()
    T.insert(pts)
    for c in T.all_cells:
        out = T.flip(c, 0)
        # assert(out == True)
    assert(T.num_edges == nedges)
    for e in T.all_edges:
        out = e.flip()
        # assert(out == True)
        del out

# Far too slow
# def test_flippable():
#     T = Delaunay3()
#     T.insert(pts)
#     for c in T.all_cells:
#         if T.flip(c, 0):
#             T.flip_flippable(c, 0)
#     assert(T.num_edges == nedges)
#     for e in T.all_edges:
#         if e.flip():
#             e.flip_flippable()


def test_io():
    fname = 'test_io2348_3.dat'
    Tout = Delaunay3()
    Tout.insert(pts)
    Tout.write_to_file(fname)
    Tin = Delaunay3()
    Tin.read_from_file(fname)
    assert(Tout.num_verts == Tin.num_verts)
    assert(Tout.num_cells == Tin.num_cells)
    os.remove(fname)


def test_vert_incident_verts():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_vertices():
            c0 += 1
            count += 1
        x = v.incident_vertices()[0]
        print(v.index, c0, x)
    print(count, 2*T.num_edges)
    assert(count == 2*T.num_edges)  # 68


def test_vert_incident_edges():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        x = v.incident_edges()[0]
        print(v.index, c0, x)
    print(count, 2*T.num_edges)
    assert(count == 2*T.num_edges)  # 68


def test_vert_incident_facets():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_facets():
            c0 += 1
            count += 1
        x = v.incident_facets()[0]
        print(v.index, c0, x)
    print(count, 3*T.num_facets)
    assert(count == 3*T.num_facets)  # 144


def test_vert_incident_cells():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_cells():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count, 4*T.num_cells)
    assert(count == 4*T.num_cells)  # 96


def test_edge_incident_verts():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count, 2*T.num_edges)
    assert(count == 2*T.num_edges)  # 68


def test_edge_incident_edges():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 404)


def test_edge_incident_facets():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_facets():
            c0 += 1
            count += 1
        print(c0)
    print(count, 3*T.num_facets)
    assert(count == 3*T.num_facets)  # 144


def test_edge_incident_cells():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count, 3*T.num_facets)
    assert(count == 3*T.num_facets)  # 144


def test_facet_incident_verts():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_facets:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count, 3*T.num_facets)
    assert(count == 3*T.num_facets)  # 144


def test_facet_incident_edges():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_facets:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count, 3*T.num_facets)
    assert(count == 3*T.num_facets)  # 144


def test_facet_incident_facets():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_facets:
        c0 = 0
        for e in v.incident_facets():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 480)


def test_facet_incident_cells():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_facets:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count, 2*T.num_facets)
    assert(count == 2*T.num_facets)  # 96


def test_cell_incident_verts():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count, 4*T.num_cells)
    assert(count == 4*T.num_cells)  # 96


def test_cell_incident_edges():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count, 6*T.num_cells)
    assert(count == 6*T.num_cells)  # 144


def test_cell_incident_facets():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_facets():
            c0 += 1
            count += 1
        print(c0)
    print(count, 4*T.num_cells)
    assert(count == 4*T.num_cells)  # 96


def test_cell_incident_cells():
    T = Delaunay3()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count, 4*T.num_cells)
    assert(count == 4*T.num_cells)  # 72


def test_nearest_vertex():
    idx_test = 8
    T = Delaunay3()
    T.insert(pts)
    v = T.nearest_vertex(pts[idx_test, :]-0.1)
    assert(v.index == idx_test)


def test_mirror():
    T = Delaunay3()
    T.insert(pts)
    for e in T.all_facets:
        e2 = T.mirror_facet(e)
        del e2
        break
    for c in T.all_cells:
        idx = 0
        i2 = T.mirror_index(c, idx)
        assert(c == c.neighbor(idx).neighbor(i2))
        v2 = T.mirror_vertex(c, idx)
        assert(v2 == c.neighbor(idx).vertex(i2))


def test_get_boundary_of_conflicts():
    T = Delaunay3()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    edges = T.get_boundary_of_conflicts(p, c)
    print(len(edges))


def test_get_conflicts():
    T = Delaunay3()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells = T.get_conflicts(p, c)
    print(len(cells))


def test_get_conflicts_and_boundary():
    T = Delaunay3()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells, edges = T.get_conflicts_and_boundary(p, c)
    print(len(cells), len(edges))


def test_vertices():
    T = Delaunay3()
    T.insert(pts)
    v = T.vertices
    assert(v.shape[0] == pts.shape[0])
    assert(v.shape[1] == pts.shape[1])
    assert(np.allclose(pts, v))


def test_edges():
    T = Delaunay3()
    T.insert(pts)
    e = T.edges
    assert(e.shape[0] == T.num_finite_edges)
    assert(e.shape[1] == 2)



r"""Tests for 2D Delaunay triangulation.

.. todo::
   * value test circumcenter
   * test includes_edge
   * answer testing for voronoi volumes
"""
import numpy as np
import os
from cgal4py.delaunay import Delaunay2
from test_cgal4py import MyTestCase, make_points


pts, left_edge, right_edge = make_points(0, 2)
pts_dup = np.concatenate([pts, np.reshape(pts[0, :], (1, pts.shape[1]))])
nverts_fin = pts.shape[0]
nverts_inf = 1
nverts = nverts_fin + nverts_inf
nedges_fin = 17
nedges_inf = 4
nedges = nedges_fin + nedges_inf
ncells_fin = 10
ncells_inf = 4
ncells = ncells_fin + ncells_inf


class TestDelaunay2(MyTestCase):

    def setup_param(self):
        self._func = Delaunay2
        self.param_runs = [
            ((pts,), {}),
            ((pts_dup,), {}),
            ]
        self.T = Delaunay2()
        self.T.insert(pts)
        self.Tdup = Delaunay2()
        self.Tdup.insert(pts_dup)
        self.pts = pts
        self.pts_dup = pts_dup

    def check_runs(self, args, kwargs):
        out = self.func()
        out.insert(*args, **kwargs)
        assert(out.is_valid())

    def new_T(self):
        T = Delaunay2()
        T.insert(pts)
        return T

    def test_equal(self):
        assert(self.T.is_equivalent(self.T))

    def check_num_verts(self, T):
        print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
        assert(T.num_finite_verts == nverts_fin)
        assert(T.num_infinite_verts == nverts_inf)
        assert(T.num_verts == nverts)

    def check_num_edges(self, T):
        print(T.num_finite_edges, T.num_infinite_edges, T.num_edges)
        assert(T.num_finite_edges == nedges_fin)
        assert(T.num_infinite_edges == nedges_inf)
        assert(T.num_edges == nedges)

    def check_num_cells(self, T):
        print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
        assert(T.num_finite_cells == ncells_fin)
        assert(T.num_infinite_cells == ncells_inf)
        assert(T.num_cells == ncells)

    def test_counts_generate(self):
        for T in [self.T, self.Tdup]:
            yield self.check_num_verts, T
            yield self.check_num_edges, T
            yield self.check_num_cells, T

    def test_all_verts(self):
        T = self.T
        count_fin = count_inf = 0
        for v in T.all_verts:
            print(v.index, v.point)
            if v.is_infinite():
                count_inf += 1
            else:
                count_fin += 1
        count = count_fin + count_inf
        assert(count_fin == T.num_finite_verts)
        assert(count_inf == T.num_infinite_verts)
        assert(count == T.num_verts)

    def test_finite_verts(self):
        T = self.T
        count = 0
        for v in T.finite_verts:
            assert((not v.is_infinite()))
            count += 1
        assert(count == T.num_finite_verts)

    def test_all_edges(self):
        T = self.T
        count_fin = count_inf = 0
        for e in T.all_edges:
            if e.is_infinite():
                count_inf += 1
            else:
                count_fin += 1
        print(count_fin, count_inf)
        count = count_fin + count_inf
        assert(count_fin == T.num_finite_edges)
        assert(count_inf == T.num_infinite_edges)
        assert(count == T.num_edges)

    def test_finite_edges(self):
        T = self.T
        count = 0
        for e in T.finite_edges:
            assert((not e.is_infinite()))
            count += 1
        print(count)
        assert(count == T.num_finite_edges)

    def test_all_cells(self):
        T = self.T
        count_fin = count_inf = 0
        for c in T.all_cells:
            if c.is_infinite():
                count_inf += 1
            else:
                count_fin += 1
        count = count_fin + count_inf
        assert(count_fin == T.num_finite_cells)
        assert(count_inf == T.num_infinite_cells)
        assert(count == T.num_cells)

    def test_finite_cells(self):
        T = self.T
        count = 0
        for c in T.finite_cells:
            assert((not c.is_infinite()))
            count += 1
        assert(count == T.num_finite_cells)

    def test_clear(self):
        T = self.new_T()
        T.clear()
        print(T.num_finite_verts, T.num_cells)
        assert(T.num_finite_verts == 0)
        assert(T.num_cells == 0)

    def test_io(self):
        fname = 'test_io2348_2.dat'
        Tout = self.T
        Tout.write_to_file(fname)
        Tin = Delaunay2.from_file(fname)
        assert(Tout.num_verts == Tin.num_verts)
        assert(Tout.num_cells == Tin.num_cells)
        os.remove(fname)

    def test_serialize(self):
        Tout = self.T
        f, n, i = Tout.serialize()
        Tin = Delaunay2.from_serial(self.pts, f, n, i)
        assert(Tout.num_verts == Tin.num_verts)
        assert(Tout.num_cells == Tin.num_cells)

    def test_locate(self):
        T = self.T
        for c in T.finite_cells:
            p = c.center
            print('{}\n{}\n{}'.format(c, p, T.locate(p)))
            assert(c == T.locate(p, c))
            assert(c == T.locate(p))
            assert(c.vertex(0) == T.locate(c.vertex(0).point))
            assert(c.vertex(0) == T.locate(c.vertex(0).point, c))
            # assert(c.edge(0) == T.locate(c.edge(0).midpoint))
            # assert(c.edge(0) == T.locate(c.edge(0).midpoint, c))
            break

    def test_get_vertex(self):
        T = self.T
        for i in range(nverts_fin):
            v = T.get_vertex(i)
            assert(np.allclose(v.point, self.pts[i, :]))

    def test_remove(self):
        T = self.new_T()
        v = T.get_vertex(0)
        T.remove(v)
        assert(T.num_verts == (nverts-1))

    def test_is_edge(self):
        T = self.T
        assert(T.is_edge(T.get_vertex(0), T.get_vertex(1)))
        assert(not T.is_edge(T.get_vertex(0), T.get_vertex(nverts_fin-1)))

    def test_is_cell(self):
        T = self.T
        assert(T.is_cell(T.get_vertex(0), T.get_vertex(1), T.get_vertex(3)))
        assert(not T.is_cell(T.get_vertex(0), T.get_vertex(1),
                             T.get_vertex(nverts_fin-1)))

    def test_vert(self):
        T = self.new_T()
        vold = None
        for v in T.all_verts:
            pnt = v.point
            idx = v.index
            vol = v.dual_volume
            print(v, idx, pnt, vol)
            assert(v == v)
            if vold is not None:
                assert(v != vold)
            if v.is_infinite():
                assert(np.isinf(pnt).all())
                assert(idx == np.iinfo(np.uint32).max)
                assert(np.isclose(vol, -1))
            else:
                assert(np.allclose(pnt, pts[idx, :]))
                if idx >= 4:
                    assert(np.isclose(vol, -1))
                c = v.cell
                v.set_cell(c)
                v.set_point(pnt)
            vold = v


# class TestVertex2(MyTestCase):

#     def setup_param(self):
#         self._func = None
#         self.pts = make_points(0, 2)
#         self.T = Delaunay2()
#         self.T.insert(self.pts)
#         self.v = self.T.finite_verts

#     def check_point(self, v):
#         pnt = v.point
#         idx = v.index
#         if v.isinfinite():
#             assert(np.isinf(pnt).all())
#             assert(idx == np.iinfo(np.uint32).max)
#         else:
#             assert(np.allclose(pnt, self.pts[idx, :]))


def test_edge():
    T = Delaunay2()
    T.insert(pts)
    eold = None
    for e in T.all_edges:
        v1 = e.vertex1
        v2 = e.vertex2
        assert(e.vertex(0) == v1)
        assert(e.vertex(1) == v2)
        elen = e.length
        print(e, v1.index, v2.index, elen, e.midpoint, e.center)
        assert(e == e)
        if eold is not None:
            assert(e != eold)
        if e.is_infinite():
            assert(np.isclose(elen, -1.0))
            assert(np.isinf(e.center).all())
        else:
            l = np.sqrt(np.sum((pts[v1.index, :]-pts[v2.index, :])**2.0))
            assert(np.isclose(elen, l))
        eold = e


def test_cell():
    T = Delaunay2()
    T.insert(pts)
    cold = None
    for c in T.all_cells:
        print(c, c.dimension, c.circumcenter, c.center)
        assert(c == c)
        if cold is not None:
            assert(c != cold)
        v1 = c.vertex(0)
        v2 = c.vertex(1)
        v3 = c.vertex(2)
        e1 = c.edge(0)
        e2 = c.edge(1)
        e3 = c.edge(2)
        assert(c.has_vertex(v1))
        assert(c.has_vertex(v1, return_index=True) == 0)
        assert(c.ind_vertex(v1) == 0)
        del(e1, e2, e3)

        c.reset_vertices()
        c.set_vertex(0, v1)
        c.set_vertices(v3, v2, v1)

        n1 = c.neighbor(0)
        n2 = c.neighbor(1)
        n3 = c.neighbor(2)
        assert(c.has_neighbor(n1))
        assert(c.has_neighbor(n1, return_index=True) == 0)
        assert(c.ind_neighbor(n1) == 0)

        c.reset_neighbors()
        c.set_neighbor(0, n1)
        c.set_neighbors(n3, n2, n1)

        print(c.side(c.center))
        print(c.side(v1.point))
        print(c.side(2*v1.point-c.center))
        print(c.side_of_circle(c.circumcenter))
        print(c.side_of_circle(v1.point))
        print(c.side_of_circle(2*v1.point-c.circumcenter),
              2*v1.point-c.circumcenter)
        if c.is_infinite():
            assert(np.isinf(c.center).all())
            assert(c.side(c.center) == -1)
            assert(c.side(v1.point) == -1)
            assert(c.side(2*v1.point-c.center) == -1)
            assert(np.isinf(c.circumcenter).all())
            assert(c.side_of_circle(c.circumcenter) == -1)
            assert(c.side_of_circle(v1.point) == -1)
            assert(c.side_of_circle(2*v1.point-c.circumcenter) == -1)
        else:
            assert(c.side(c.center) == -1)
            assert(c.side(v1.point) == 0)
            assert(c.side(2*v1.point-c.center) == 1)
            assert(c.side_of_circle(c.circumcenter) == -1)
            assert(c.side_of_circle(v1.point) == 0)
            assert(c.side_of_circle(2*v1.point-c.circumcenter) == 1)

        c.reorient()
        c.ccw_permute()
        c.cw_permute()

        cold = c


def test_move():
    T = Delaunay2()
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(2, 'float64')
    v = T.move(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(T.num_verts == (nverts-1))


def test_move_if_no_collision():
    T = Delaunay2()
    T.insert(pts)
    v0 = T.get_vertex(0)
    new_pos = np.zeros(2, 'float64')
    v = T.move_if_no_collision(v0, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v0.point, new_pos))
    v1 = T.get_vertex(1)
    v = T.move_if_no_collision(v1, new_pos)
    assert(np.allclose(v.point, new_pos))
    assert(np.allclose(v1.point, pts[1, :]))
    assert(T.num_verts == nverts)


def test_flip():
    T = Delaunay2()
    T.insert(pts)
    for c in T.all_cells:
        out = T.flip(c, 0)
        assert(out)
    print(T.num_edges, nedges)
    assert(T.num_edges == nedges)
    for e in T.all_edges:
        out = e.flip()
        assert(out)
    print(T.num_edges, nedges)


def test_flippable():
    T = Delaunay2()
    T.insert(pts)
    for c in T.all_cells:
        T.flip(c, 0)
    print(T.num_edges, nedges)
    assert(T.num_edges == nedges)
    for e in T.all_edges:
        e.flip()
    print(T.num_edges, nedges)


def test_vert_incident_verts():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_vertices():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 42)


def test_vert_incident_edges():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 42)


def test_vert_incident_cells():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_cells():
            c0 += 1
            count += 1
        print(v.index, c0)
    print(count)
    assert(count == 42)


def test_edge_incident_verts():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 42)


def test_edge_incident_edges():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 156)


def test_edge_incident_cells():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_edges:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 100)


def test_cell_incident_verts():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 42)


def test_cell_incident_edges():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_edges():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 42)


def test_cell_incident_cells():
    T = Delaunay2()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count)
    assert(count == 42)


def test_nearest_vertex():
    idx_test = 7
    T = Delaunay2()
    T.insert(pts)
    v = T.nearest_vertex(pts[idx_test, :]-0.1)
    assert(v.index == idx_test)


def test_mirror():
    T = Delaunay2()
    T.insert(pts)
    for e in T.all_edges:
        e2 = T.mirror_edge(e)
        del(e2)
        break
    for c in T.all_cells:
        idx = 0
        i2 = T.mirror_index(c, idx)
        assert(c == c.neighbor(idx).neighbor(i2))
        v2 = T.mirror_vertex(c, idx)
        assert(v2 == c.neighbor(idx).vertex(i2))


def test_get_boundary_of_conflicts():
    T = Delaunay2()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    edges = T.get_boundary_of_conflicts(p, c)
    print(len(edges))


def test_get_conflicts():
    T = Delaunay2()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells = T.get_conflicts(p, c)
    print(len(cells))


def test_get_conflicts_and_boundary():
    T = Delaunay2()
    T.insert(pts)
    v = T.get_vertex(0)
    c = v.incident_cells()[0]
    p = c.circumcenter
    cells, edges = T.get_conflicts_and_boundary(p, c)
    print(len(cells), len(edges))


def test_line_walk():
    T = Delaunay2()
    T.insert(pts)
    p1 = np.array([-1, -1], 'float64')
    p2 = np.array([+1, +1], 'float64')
    x = T.line_walk(p1, p2)
    print(len(x))
    assert(len(x) == 6)


def test_vertices():
    T = Delaunay2()
    T.insert(pts)
    v = T.vertices
    assert(v.shape[0] == pts.shape[0])
    assert(v.shape[1] == pts.shape[1])
    assert(np.allclose(pts, v))


def test_edges():
    T = Delaunay2()
    T.insert(pts)
    e = T.edges
    assert(e.shape[0] == T.num_finite_edges)
    assert(e.shape[1] == 2)


def test_boundary_points():
    T = Delaunay2()
    T.insert(pts)
    out = T.boundary_points(left_edge, right_edge, False)
    del(out)


def test_outgoing_points():
    T = Delaunay2()
    T.insert(pts)
    le = np.array([[-2, -2],
                   [-1, -1],
                   [+0,  0],
                   [+1,  1]], 'float64')
    re = np.array([[-1, -1],
                   [+0,  0],
                   [+1,  1],
                   [+2,  2]], 'float64')
    out = T.outgoing_points(le, re)
    assert(len(out) == le.shape[0])


def test_voronoi_volumes():
    T = Delaunay2()
    T.insert(pts)
    v = T.voronoi_volumes()
    assert(v.shape[0] == T.num_finite_verts)



from cgal4py.tests.test_cgal4py import make_points
from cgal4py.delaunay import _get_Delaunay
from cgal4py import triangulate, voronoi_volumes
import sys
from datetime import datetime

from mpi4py import MPI
import numpy as np

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

limit_mem = 32
use_double = True

periodic = True
npts = int(1e9)
ndim = 3
test_result = False
task = 'vols'
unique_str = datetime.today().strftime("%Y%j%H%M%S")

ParallelDelaunay = _get_Delaunay(ndim, bit64=use_double, parallel=True,
                                 comm=comm)

if rank == 0:
    pts, le, re = make_points(npts, ndim)
    pts2, le, re = make_points(10, ndim)
else:
    pts = None
    le = None
    re = None
    pts2 = None

TP = ParallelDelaunay(le, re, periodic=periodic, limit_mem=limit_mem,
                      unique_str=unique_str)
TP.insert(pts)
# TP.insert(pts2)
if rank == 0:
    print("Consolidating")
if task == 'tess':
    T_new = TP.consolidate_tess()
else:
    V_new = TP.consolidate_vols()
if rank == 0:
    if test_result:
        if task == 'tess':
            T_old = triangulate(pts, le, re, nproc=1)#size)
            c_old, n_old, inf_old = T_old.serialize(sort=True)
            c_new, n_new, inf_new = T_new.serialize(sort=True)
            try:
                assert(np.all(c_old == c_new))
                assert(np.all(n_old == n_new))
                assert(T_new.is_equivalent(T_old))
            except:
                for name, T in zip(['New','Old'],[T_new, T_old]):
                    print(name)
                    print('\t verts: {}, {}, {}'.format(
                        T.num_verts, T.num_finite_verts, T.num_infinite_verts))
                    print('\t cells: {}, {}, {}'.format(
                        T.num_cells, T.num_finite_cells, T.num_infinite_cells))
                    print('\t edges: {}, {}, {}'.format(
                        T.num_edges, T.num_finite_edges, T.num_infinite_edges))
                    if ndim == 3:
                        print('\t facets: {}, {}, {}'.format(
                            T.num_facets, T.num_finite_facets,
                            T.num_infinite_facets))
                raise
        else:
            V_old = voronoi_volumes(pts, le, re, nproc=1)
            assert(np.allclose(V_new, V_old))
    print("Success!!")



r"""Tests for nD Delaunay Triangulation.

.. todo:
   * answer test circumcenter
   * answer test ability to flip
   * clarify equality between facets defined using different cells

"""
import numpy as np
import os
import itertools
from cgal4py.delaunay import _get_Delaunay
from nose.tools import nottest

ndim = 4
DelaunayD = _get_Delaunay(ndim, overwrite=False)

left_edge = -2*np.ones(ndim, 'float64')
right_edge = 2*np.ones(ndim, 'float64')
pts = np.array([[0 for _ in range(ndim)]]+
               [i for i in itertools.product([-1, 1], repeat=ndim)],
               'float64')
pts[-1,-1] += 0.0000001
pts_dup = np.concatenate([pts, np.reshape(pts[0, :], (1, pts.shape[1]))])
nverts_fin = pts.shape[0]
nverts_inf = 1
nverts = nverts_fin + nverts_inf
if ndim == 4:
    ncells_fin = 51
    ncells_inf = 51
    cvol = 11.33333333333
elif ndim == 5:
    ncells_fin = 260
    ncells_inf = 260
    cvol = 24.739583201584463
else:
    ncells_fin = 0
    ncells_inf = 0
    cvol = 0.0
ncells = ncells_fin + ncells_inf


@nottest
def count_faces_per_cell(face_dim):
    N = ndim+1
    K = face_dim+1
    return np.math.factorial(N)/(np.math.factorial(N-K)*np.math.factorial(K))


def test_create():
    T = DelaunayD()
    del T


def test_insert():
    # without duplicates
    T = DelaunayD()
    T.insert(pts)
    assert(T.is_valid())
    # with duplicates
    T = DelaunayD()
    T.insert(pts_dup)
    assert(T.is_valid())


def test_equal():
    T1 = DelaunayD()
    T1.insert(pts)
    T2 = DelaunayD()
    T2.insert(pts)
    assert(T1.is_equivalent(T2))


def test_num_verts():
    # without duplicates
    T = DelaunayD()
    T.insert(pts)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)
    # with duplicates
    T = DelaunayD()
    T.insert(pts_dup)
    print(T.num_finite_verts, T.num_infinite_verts, T.num_verts)
    assert(T.num_finite_verts == nverts_fin)
    assert(T.num_infinite_verts == nverts_inf)
    assert(T.num_verts == nverts)


def test_num_cells():
    # without duplicates
    T = DelaunayD()
    T.insert(pts)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)
    # with duplicates
    T = DelaunayD()
    T.insert(pts_dup)
    print(T.num_finite_cells, T.num_infinite_cells, T.num_cells)
    assert(T.num_finite_cells == ncells_fin)
    assert(T.num_infinite_cells == ncells_inf)
    assert(T.num_cells == ncells)


def test_all_verts():
    T = DelaunayD()
    T.insert(pts)
    count_fin = count_inf = 0
    for v in T.all_verts:
        if v.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_finite_verts)
    assert(count_inf == T.num_infinite_verts)
    assert(count == T.num_verts)


def test_finite_verts():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.finite_verts:
        assert((not v.is_infinite()))
        count += 1
    assert(count == T.num_finite_verts)


def test_all_cells():
    T = DelaunayD()
    T.insert(pts)
    count_fin = count_inf = 0
    for c in T.all_cells:
        if c.is_infinite():
            count_inf += 1
        else:
            count_fin += 1
    count = count_fin + count_inf
    assert(count_fin == T.num_finite_cells)
    assert(count_inf == T.num_infinite_cells)
    assert(count == T.num_cells)


def test_finite_cells():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for c in T.finite_cells:
        assert((not c.is_infinite()))
        count += 1
    assert(count == T.num_finite_cells)


def test_get_vertex():
    T = DelaunayD()
    T.insert(pts)
    for i in range(nverts_fin):
        v = T.get_vertex(i)
        assert(np.allclose(v.point, pts[i, :]))


def test_locate():
    T = DelaunayD()
    T.insert(pts)
    for c in T.finite_cells:
        r_cell = c
        p_cell = r_cell.center
        print(r_cell, p_cell, T.locate(p_cell))
        assert(r_cell == T.locate(p_cell))
        assert(r_cell == T.locate(p_cell, c))
        r_vert = c.vertex(0)
        p_vert = r_vert.point
        print(r_vert, p_vert, T.locate(p_vert))
        assert(r_vert == T.locate(p_vert))
        assert(r_vert == T.locate(p_vert, c))
        r_facet = c.facet(0)
        p_facet = r_facet.center
        # TODO: non-equivalence of facets with same vertices
        # x = T.locate(p_facet)
        # print("facet")
        # for i in range(r_facet.nverts):
        #     print(r_facet.vertex(i))
        # print("located")
        # for i in range(x.nverts):
        #     print(x.vertex(i))
        # print(r_facet, p_facet, T.locate(p_facet))
        # assert(r_facet.is_equivalent(r_facet))
        # assert(r_facet.is_equivalent(T.locate(p_facet)))
        # assert(r_facet.is_equivalent(T.locate(p_facet, c)))
        # assert(r_facet == T.locate(p_facet))
        # assert(r_facet == T.locate(p_facet, c))
        break


def test_remove():
    T = DelaunayD()
    T.insert(pts)
    v = T.get_vertex(0)
    T.remove(v)
    assert(T.num_verts == (nverts-1))


def test_clear():
    T = DelaunayD()
    T.insert(pts)
    T.clear()
    print(T.num_finite_verts, T.num_cells)
    assert(T.num_finite_verts == 0)
    assert(T.num_cells == 1)


def test_vert():
    T = DelaunayD()
    T.insert(pts)
    vold = None
    for v in T.all_verts:
        idx = v.index
        pnt = v.point
        vol = v.dual_volume
        print(v, idx, pnt, vol)
        assert(v == v)
        if vold is not None:
            assert(v != vold)
        if v.is_infinite():
            assert(idx == np.iinfo(np.uint32).max)
            assert(np.isinf(pnt).all())
            assert(np.isclose(vol, -1.0))
        else:
            assert(np.allclose(pnt, pts[idx, :]))
            if idx == 0:
                assert(np.isclose(vol, cvol))
            else:
                assert(np.isclose(vol, -1.0))
            c = v.cell
            v.set_cell(c)
            v.set_point(pnt)
        vold = v


# def test_edge():
#     T = DelaunayD()
#     T.insert(pts)
#     eold = None
#     for e in T.all_edges:
#         v1 = e.vertex(0)
#         v2 = e.vertex(1)
#         assert(v1 == e.vertex1)
#         assert(v2 == e.vertex2)
#         c = e.cell
#         i1 = e.ind1
#         i2 = e.ind2
#         elen = e.length
#         inf = e.is_infinite()
#         gab = e.is_Gabriel()
#         print(e, v1.index, v2.index, elen, inf, gab)
#         assert(e == e)
#         assert(e.is_equivalent(e))
#         if eold is not None:
#             assert(e != eold)
#         p1 = e.center
#         p2 = v1.point
#         print(e.side(p1), p1)
#         print(e.side(p2), p2)
#         if e.is_infinite():
#             assert(np.isclose(elen, -1.0))
#             assert(np.isinf(e.center).all())
#             assert(e.side(p1) == -1)
#             assert(e.side(p2) == -1)
#         else:
#             l = np.sqrt(np.sum((pts[v1.index, :]-pts[v2.index, :])**2.0))
#             assert(np.isclose(elen, l))
#             # p3 = e.center + 10*elen
#             # print(e.side(p3), p3)
#             # assert(e.side(p1) == -1) # virtually impossible
#             # assert(e.side(p2) == 0)
#             # assert(e.side(p3) == 1)
#         eold = e
#         del(c, i1, i2)


# def test_facet():
#     T = DelaunayD()
#     T.insert(pts)
#     fold = None
#     for f in T.all_facets:
#         v1 = f.vertex(0)
#         v2 = f.vertex(1)
#         v3 = f.vertex(2)
#         e1 = f.edge(0)
#         e2 = f.edge(1)
#         e3 = f.edge(2)
#         c = f.cell
#         i = f.ind
#         inf = f.is_infinite()
#         gab = f.is_Gabriel()
#         print(f, v1.index, v2.index, v3.index, i, inf, gab, f.center)
#         assert(f == f)
#         assert(f.is_equivalent(f))
#         if fold is not None:
#             assert(f != fold)
#         del(e1, e2, e3, c)

#         p1 = f.center
#         p2 = v1.point
#         print(f.side(p1), p1)
#         print(f.side(p2), p2)
#         if f.is_infinite():
#             assert(np.isinf(f.center).all())
#             assert(f.side(p1) == -1)
#             assert(f.side(p2) == -1)
#         # else:
#         #     p3 = 2*v1.point - f.center + np.arange(3)
#         #     print(f.side(p3), p3)
#         #     assert(f.side(p1) == -1)
#         #     assert(f.side(p2) == 0)
#         #     assert(f.side(p3) == 1)

#         # # This segfaults inside CGAL function call
#         # print(f.side_of_circle((v1.point+v2.point+v3.point)/3),
#         #       (v1.point+v2.point+v3.point)/3)
#         # print(f.side_of_circle(v1.point), v1.point)
#         # print(f.side_of_circle((5*v1.point-v2.point-v3.point)/3),
#         #       (5*v1.point-v2.point-v3.point)/3)
#         # if f.is_infinite():
#         #     assert(f.side_of_circle((v1.point+v2.point+v3.point)/3) == -1)
#         #     assert(f.side_of_circle(v1.point) == -1)
#         #     assert(f.side_of_circle((5*v1.point-v2.point-v3.point)/3) == -1)
#         # else:
#         #     # This segfaults...
#         #     assert(f.side_of_circle((v1.point+v2.point+v3.point)/3) == -1)
#         #     assert(f.side_of_circle(v1.point) == 0)
#         #     assert(f.side_of_circle((5*v1.point-v2.point-v3.point)/3) == 1)

#         fold = f


def test_cell():
    T = DelaunayD()
    T.insert(pts)
    cold = None
    for c in T.all_cells:
        print(c, c.circumcenter, c.center)
        assert(c == c)
        if cold is not None:
            assert(c != cold)

        f1 = c.facet(0)
        del(f1)

        v1 = c.vertex(0)
        assert(c.has_vertex(v1))
        assert(c.has_vertex(v1, return_index=True) == 0)
        assert(c.ind_vertex(v1) == 0)

        c.set_vertex(0, v1)

        n1 = c.neighbor(0)
        assert(c.has_neighbor(n1))
        assert(c.has_neighbor(n1, return_index=True) == 0)
        assert(c.ind_neighbor(n1) == 0)

        c.set_neighbor(0, n1)

        cold = c


# def test_io():
#     fname = 'test_io2348_3.dat'
#     Tout = DelaunayD()
#     Tout.insert(pts)
#     Tout.write_to_file(fname)
#     Tin = DelaunayD()
#     Tin.read_from_file(fname)
#     assert(Tout.num_verts == Tin.num_verts)
#     assert(Tout.num_cells == Tin.num_cells)
#     os.remove(fname)


def test_vert_incident_verts():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_vertices():
            c0 += 1
            count += 1
        x = v.incident_vertices()[0]
        print(v.index, c0, x)
    # print(count, 2*T.num_edges)
    # assert(count == 2*T.num_edges)


def test_vert_incident_edges():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_faces(1):
            c0 += 1
            count += 1
        x = v.incident_faces(1)[0]
        print(v.index, c0, x)
    # print(count, 2*T.num_edges)
    # assert(count == 2*T.num_edges)


def test_vert_incident_facets():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for e in v.incident_faces(ndim-1):
            c0 += 1
            count += 1
        x = v.incident_faces(ndim-1)[0]
        print(v.index, c0, x)
    # print(count, (ndim)*T.num_facets)
    # assert(count == (ndim)*T.num_facets)


def test_vert_incident_cells():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_verts:
        c0 = 0
        for c in v.incident_cells():
            c0 += 1
            count += 1
        print(v.index, c0)
    expected = count_faces_per_cell(0)*T.num_cells
    print(count, expected)
    assert(count == expected)


# def test_edge_incident_verts():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_edges:
#         c0 = 0
#         for e in v.incident_vertices():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 2*T.num_edges)
#     assert(count == 2*T.num_edges)  # 68


# def test_edge_incident_edges():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_edges:
#         c0 = 0
#         for e in v.incident_edges():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count)
#     assert(count == 404)


# def test_edge_incident_facets():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_edges:
#         c0 = 0
#         for e in v.incident_facets():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 3*T.num_facets)
#     assert(count == 3*T.num_facets)  # 144


# def test_edge_incident_cells():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_edges:
#         c0 = 0
#         for e in v.incident_cells():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 3*T.num_facets)
#     assert(count == 3*T.num_facets)  # 144


# def test_facet_incident_verts():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_facets:
#         c0 = 0
#         for e in v.incident_vertices():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 3*T.num_facets)
#     assert(count == 3*T.num_facets)  # 144


# def test_facet_incident_edges():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_facets:
#         c0 = 0
#         for e in v.incident_edges():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 3*T.num_facets)
#     assert(count == 3*T.num_facets)  # 144


# def test_facet_incident_facets():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_facets:
#         c0 = 0
#         for e in v.incident_facets():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count)
#     assert(count == 480)


# def test_facet_incident_cells():
#     T = DelaunayD()
#     T.insert(pts)
#     count = 0
#     for v in T.all_facets:
#         c0 = 0
#         for e in v.incident_cells():
#             c0 += 1
#             count += 1
#         print(c0)
#     print(count, 2*T.num_facets)
#     assert(count == 2*T.num_facets)  # 96


def test_cell_incident_verts():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_vertices():
            c0 += 1
            count += 1
        print(c0)
    expected = count_faces_per_cell(0)*T.num_cells
    print(count, expected)
    assert(count == expected)


def test_cell_incident_edges():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_faces(1):
            c0 += 1
            count += 1
        print(c0)
    expected = count_faces_per_cell(1)*T.num_cells
    print(count, expected)
    assert(count == expected)


def test_cell_incident_facets():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_faces(ndim-1):
            c0 += 1
            count += 1
        print(c0)
    expected = count_faces_per_cell(ndim-1)*T.num_cells
    print(count, expected)
    assert(count == expected)

def test_cell_incident_cells():
    T = DelaunayD()
    T.insert(pts)
    count = 0
    for v in T.all_cells:
        c0 = 0
        for e in v.incident_cells():
            c0 += 1
            count += 1
        print(c0)
    print(count, (ndim+1)*T.num_cells)
    assert(count == (ndim+1)*T.num_cells)


def test_mirror():
    T = DelaunayD()
    T.insert(pts)
    for c in T.all_cells:
        idx = 0
        i2 = T.mirror_index(c, idx)
        assert(c == c.neighbor(idx).neighbor(i2))
        v2 = T.mirror_vertex(c, idx)
        assert(v2 == c.neighbor(idx).vertex(i2))


def test_vertices():
    T = DelaunayD()
    T.insert(pts)
    v = T.vertices
    assert(v.shape[0] == pts.shape[0])
    assert(v.shape[1] == pts.shape[1])
    assert(np.allclose(pts, v))



cimport numpy as np
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from libcpp.pair cimport pair
from libcpp cimport bool
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t

cdef extern from "c_periodic_delaunay2.hpp":
    cdef int VALID

    cdef cppclass PeriodicDelaunay_with_info_2[Info] nogil:
        PeriodicDelaunay_with_info_2() except +
        PeriodicDelaunay_with_info_2(const double *domain) except +
        PeriodicDelaunay_with_info_2(double *pts, Info *val, uint32_t n) except +
        PeriodicDelaunay_with_info_2(double *pts, Info *val, uint32_t n,
                                     const double *domain) except +
        bool updated
        bool is_valid() const
        void num_sheets(int32_t *ns_out) const
        uint32_t num_sheets_total() const
        uint32_t num_finite_verts() const
        uint32_t num_finite_edges() const
        uint32_t num_finite_cells() const
        uint32_t num_infinite_verts() const
        uint32_t num_infinite_edges() const
        uint32_t num_infinite_cells() const
        uint32_t num_verts() const
        uint32_t num_edges() const
        uint32_t num_cells() const
        uint32_t num_stored_verts() const
        uint32_t num_stored_edges() const
        uint32_t num_stored_cells() const

        bool is_equal(const PeriodicDelaunay_with_info_2[Info] other) const

        cppclass Vertex
        cppclass Edge
        cppclass Cell

        bool has_offset(Vertex v) const
        bool has_offset(Edge e) const
        bool has_offset(Cell c) const
        void point(Vertex v, double* pos) const
        void periodic_point(Vertex v, double* pos) const
        void periodic_offset(Vertex v, int32_t* off) const
        void point(Edge e, int i, double* pos) const
        void periodic_point(Edge e, int i, double* pos) const
        void periodic_offset(Edge e, int i, int32_t* off) const
        void point(Cell c, int i, double* pos) const
        void periodic_point(Cell c, int i, double* pos) const
        void periodic_offset(Cell c, int i, int32_t* off) const

        void set_domain(const double *domain) except +
        void insert(double *, Info *val, uint32_t n) except +
        void remove(Vertex v) except +
        void clear() except +
        Vertex move(Vertex v, double *pos) except +
        Vertex move_if_no_collision(Vertex v, double *pos) except +

        void write_to_file(const char* filename) except +
        void read_from_file(const char* filename) except +
        I serialize[I](I &n, I &m, int32_t &d,
                       double* domain, int32_t* cover,
                       double* vert_pos, Info* vert_info,
                       I* faces, I* neighbors, int32_t *offsets) const
        Info serialize_idxinfo[I](I &n, I &m, int32_t &d, 
                                  double* domain, int32_t* cover,
                                  Info* faces, I* neighbors, int32_t *offsets) const
        I serialize_info2idx[I](I &n, I &m, int32_t &d,
                                double* domain, int32_t* cover,
                                I* faces, I* neighbors, int32_t *offsets,
                                Info max_info, I* idx) const
        void deserialize[I](I n, I m, int32_t d,
                            double* domain, int32_t *cover,
                            double* vert_pos, Info* vert_info,
                            I* faces, I* neighbors, int32_t* offsets, I idx_inf) 
        void deserialize_idxinfo[I](I n, I m, int32_t d, 
                                    double* domain, int32_t *cover,
                                    double* vert_pos,
                                    I* faces, I* neighbors, int32_t* offsets, I idx_inf)

        Vertex get_vertex(Info index) except +
        Cell locate(double* pos, int& lt, int& li)
        Cell locate(double* pos, int& lt, int& li, Cell c)

        void info_ordered_vertices(double* pos)
        void vertex_info(Info* verts)
        void edge_info(Info* edges)

        cppclass All_verts_iter:
            All_verts_iter()
            All_verts_iter& operator++()
            All_verts_iter& operator--()
            bool operator==(All_verts_iter other)
            bool operator!=(All_verts_iter other)
        All_verts_iter all_verts_begin()
        All_verts_iter all_verts_end()

        cppclass Vertex:
            Vertex()
            Vertex(All_verts_iter v)
            bool operator==(Vertex other)
            bool operator!=(Vertex other)
            void point(double* out)
            void offset(int32_t* out)
            Info info()
            Cell cell()
            void set_cell(Cell c)
            void set_point(double* x)

        cppclass All_edges_iter:
            All_edges_iter()
            All_edges_iter& operator++()
            All_edges_iter& operator--()
            bool operator==(All_edges_iter other)
            bool operator!=(All_edges_iter other)
        All_edges_iter all_edges_begin()
        All_edges_iter all_edges_end()

        cppclass Edge:
            Edge()
            Edge(All_edges_iter it)
            Edge(Vertex x1, Vertex x2)
            Edge(Cell x, int i)
            Cell cell()
            int ind()
            Vertex v1()
            Vertex v2()
            Vertex vertex(int i)
            bool operator==(Edge other)
            bool operator!=(Edge other)

        cppclass All_cells_iter:
            All_cells_iter()
            All_cells_iter& operator++()
            All_cells_iter& operator--()
            bool operator==(All_cells_iter other)
            bool operator!=(All_cells_iter other)
        All_cells_iter all_cells_begin()
        All_cells_iter all_cells_end()

        cppclass Cell:
            Cell()
            Cell(All_cells_iter c)
            Cell(Vertex v1, Vertex v2, Vertex v3)
            Cell(Vertex v1, Vertex v2, Vertex v3, Cell c1, Cell c2, Cell c3)
            bool operator==(Cell other)
            bool operator!=(Cell other)

            Vertex vertex(int i)
            bool has_vertex(Vertex v)
            bool has_vertex(Vertex v, int *i)
            int ind(Vertex v)

            Cell neighbor(int i)
            bool has_neighbor(Cell c)
            bool has_neighbor(Cell c, int *i)
            int ind(Cell c)

            void set_vertex(int i, Vertex v)
            void set_vertices()
            void set_vertices(Vertex v1, Vertex v2, Vertex v3)
            void set_neighbor(int i, Cell c)
            void set_neighbors()
            void set_neighbors(Cell c1, Cell c2, Cell c3)

            void reorient()
            void ccw_permute()
            void cw_permute()

            int dimension() except + 

        bool is_infinite(Vertex x)
        bool is_infinite(Edge x)
        bool is_infinite(Cell x)
        bool is_infinite(All_verts_iter x)
        bool is_infinite(All_edges_iter x)
        bool is_infinite(All_cells_iter x)

        bool is_edge(Vertex x1, Vertex x2)
        bool is_edge(Vertex x1, Vertex x2, Cell& c, int& i)
        bool is_cell(Vertex x1, Vertex x2, Vertex x3)
        bool is_cell(Vertex x1, Vertex x2, Vertex x3, Cell& c)

        vector[Vertex] incident_vertices(Vertex x)
        vector[Edge] incident_edges(Vertex x)
        vector[Cell] incident_cells(Vertex x)

        vector[Vertex] incident_vertices(Edge x)
        vector[Edge] incident_edges(Edge x)
        vector[Cell] incident_cells(Edge x)

        vector[Vertex] incident_vertices(Cell x)
        vector[Edge] incident_edges(Cell x)
        vector[Cell] incident_cells(Cell x)

        Vertex nearest_vertex(double* pos)

        int mirror_index(Cell x, int i) const 
        Vertex mirror_vertex(Cell x, int i) const 

        void circumcenter(Cell x, double* out)
        double dual_area(const Vertex v)
        void dual_areas(double* vols) const
        double length(const Edge e)

        bool flip(Cell x, int i)
        bool flip(Edge x)
        void flip_flippable(Cell x, int i)
        void flip_flippable(Edge x)

        vector[Edge] get_boundary_of_conflicts(double* pos, Cell start)
        vector[Cell] get_conflicts(double* pos, Cell start)
        pair[vector[Cell],vector[Edge]] get_conflicts_and_boundary(double* pos, Cell start)

        int oriented_side(Cell f, const double* pos) const
        int side_of_oriented_circle(Cell f, const double* pos) const

        vector[vector[Info]] outgoing_points(uint64_t nbox,
                                             double *left_edges, double *right_edges)
        void boundary_points(double *left_edge, double *right_edge, bool periodic,
                             vector[Info]& lx, vector[Info]& ly,
                             vector[Info]& rx, vector[Info]& ry,
                             vector[Info]& alln) const




r"""Tests for package level methods."""
import numpy as np
from datetime import datetime
import cProfile
import pstats
import time
import signal
import nose.tools as nt
from cgal4py import triangulate, voronoi_volumes, domain_decomp


@nt.nottest
def signal_print_traceback(signo, frame):
    print(traceback.print_stack(frame))


@nt.nottest
def make_points(npts, ndim, distrib='uniform', seed=0):
    npts = int(npts)
    if npts <= 0:
        if ndim == 2:
            left_edge = -2*np.ones(2, 'float64')
            right_edge = 2*np.ones(2, 'float64')
            pts = np.array([[-0.49419885869540180, -0.07594397977563715],
                            [-0.06448037284989526,  0.49582484963658130],
                            [+0.49111543670946320,  0.09383830681375946],
                            [-0.34835358086909700, -0.35867782576523670],
                            [-1,     -1],
                            [-1,      1],
                            [+1,     -1],
                            [+1,      1]], 'float64')
        elif ndim == 3:
            left_edge = -2*np.ones(3, 'float64')
            right_edge = 2*np.ones(3, 'float64')
            pts = np.array([[+0,  0,  0],
                            [-1, -1, -1],
                            [-1, -1,  1],
                            [-1,  1, -1],
                            [-1,  1,  1],
                            [+1, -1, -1],
                            [+1, -1,  1],
                            [+1,  1, -1],
                            [+1,  1,  1.0000001]], 'float64')
        else:
            raise ValueError("Invalid 'ndim': {}".format(ndim))
        npts = pts.shape[0]
    else:
        np.random.seed(seed)
        LE = 0.0
        RE = 1.0
        left_edge = LE*np.ones(ndim, 'float64')
        right_edge = RE*np.ones(ndim, 'float64')
        if distrib == 'uniform':
            pts = np.random.uniform(low=LE, high=RE, size=(npts, ndim))
        elif distrib in ('gaussian', 'normal'):
            pts = np.random.normal(loc=(LE+RE)/2.0, scale=(RE-LE)/4.0,
                                   size=(npts, ndim))
            np.clip(pts, LE, RE)
        elif distrib in (2, '2'):
            pts = make_points(0, 2)[0]
        elif distrib in (3, '3'):
            pts = make_points(0, 3)[0]
        else:
            raise ValueError("Invalid 'distrib': {}".format(distrib))
    return pts, left_edge, right_edge


@nt.nottest
def make_test(npts, ndim, distrib='uniform', periodic=False,
              leafsize=None, nleaves=0):
    # Points
    pts, left_edge, right_edge = make_points(npts, ndim, distrib=distrib)
    npts = pts.shape[0]
    # Tree
    if leafsize is None:
        leafsize = npts/2 + 2
    tree = domain_decomp.tree("kdtree", pts, left_edge, right_edge,
                              periodic=periodic, leafsize=leafsize,
                              nleaves=nleaves)
    return pts, tree


@nt.nottest
def run_test(npts, ndim, nproc=0, func_name='tess', distrib='uniform',
             profile=False, use_mpi=False, use_python=False, use_buffer=False,
             limit_mem=False, suppress_final_output=False, **kwargs):
    r"""Run a rountine with a designated number of points & dimensions on a
    selected number of processors.

    Args:
        func (str): Name of the function that should be run. Values include:
            'Delaunay': Full triangulation.
            'VoronoiVolumes': Cell volumes from triangulation.
        npart (int): Number of particles.
        nproc (int): Number of processors.
        ndim (int): Number of dimensions.
        periodic (bool, optional): If True, the domain is assumed to be
            periodic. Defaults to False.
        use_mpi (bool, optional): If True, the MPI parallelized version is used
            instead of the version using the multiprocessing package. Defaults
            to False.
        use_python (bool, optional): If True and `use_mpi == True`, then
            communications are done in python using mpi4py. Defaults to False.
        use_buffer (bool, optional): If True and `use_mpi == True`, then buffer
            versions of MPI communications will be used rather than indirect
            communications of python objects via pickling. Defaults to False.
        limit_mem (bool, optional): If True, memory usage is limited by
            writing things to file at a cost to performance. Defaults to
            False.
        suppress_final_output (bool, optional): If True, the final output
            from spawned MPI processes is suppressed. This is mainly for
            timing purposes. Defaults to False.

    Raises:
        ValueError: If `func` is not one of the supported values.

    """
    unique_str = datetime.today().strftime("%Y%j%H%M%S")
    pts, left_edge, right_edge = make_points(npts, ndim, distrib=distrib)
    if func_name.lower() in ['tess', 'delaunay', 'triangulate']:
        func = triangulate
    elif func_name.lower() in ['vols', 'volumes', 'voronoivolumes']:
        func = voronoi_volumes
    else:
        raise ValueError("Unsupported function: {}".format(func_name))
    # Set keywords for multiprocessing version
    if nproc > 1:
        kwargs['use_mpi'] = use_mpi
        kwargs['limit_mem'] = limit_mem
        if use_mpi:
            kwargs['use_python'] = use_python
            kwargs['use_buffer'] = use_buffer
            kwargs['suppress_final_output'] = suppress_final_output
            if profile:
                kwargs['profile'] = '{}_mpi_profile.dat'.format(unique_str)
    # Run
    if profile:
        pr = cProfile.Profile()
        t0 = time.time()
        pr.enable()
    out = func(pts, left_edge=left_edge, right_edge=right_edge,
               nproc=nproc, **kwargs)
    if profile:
        pr.disable()
        t1 = time.time()
        ps = pstats.Stats(pr)
        if kwargs.get('use_mpi', False):
            ps.add(kwargs['profile'])
        if isinstance(profile, str):
            ps.dump_stats(profile)
            print("Stats saved to {}".format(profile))
        else:
            sort_key = 'tottime'
            ps.sort_stats(sort_key).print_stats(25)
            # ps.sort_stats(sort_key).print_callers(5)
            print("{} s according to 'time'".format(t1-t0))
        return ps


class MyTestCase(object):

    def __init__(self):
        self._func = None
        self.param_runs = []
        self.param_returns = []
        self.param_raises = []
        self.setup_param()

    def setup_param(self):
        pass

    @property
    def func(self):
        if self._func is None:
            raise AttributeError("_func must be set.")
        else:
            return self._func

    def check_runs(self, args, kwargs):
        self.func(*args, **kwargs)

    def check_returns(self, result, args, kwargs):
        nt.eq_(result, self.func(*args, **kwargs))

    def check_raises(self, excpt, args, kwargs):
        nt.assert_raises(excpt, self.func, *args, **kwargs)

    def test_runs_generator(self):
        for args, kwargs in self.param_runs:
            yield self.check_runs, args, kwargs

    def test_returns_generator(self):
        for res, args, kwargs in self.param_returns:
            yield self.check_returns, res, args, kwargs

    def test_raises_generator(self):
        for err, args, kwargs in self.param_raises:
            yield self.check_raises, err, args, kwargs



class TestTriangulate(MyTestCase):

    def setup_param(self):
        self._func = triangulate
        self.param_runs = []
        for ndim in [2, 3]:  # , 4]:
            pts, le, re = make_points(0, ndim)
            self.param_runs += [
                ((pts,), {}),
                ((pts,), {'periodic': True,
                          'left_edge': le, 'right_edge': re}),
                ]
        pts, le, re = make_points(0, 2)
        self.param_raises = [
            (ValueError, (np.zeros((3, 3, 3)),), {}),
            (ValueError, (pts,), {'left_edge': np.zeros(3)}),
            (ValueError, (pts,), {'right_edge': np.zeros(3)}),
            (ValueError, (pts,), {'left_edge': np.zeros((2, 2, 2))}),
            (ValueError, (pts,), {'right_edge': np.zeros((2, 2, 2))}),
            ]


class TestVoronoiVolumes(MyTestCase):

    def setup_param(self):
        self._func = voronoi_volumes
        self.param_runs = []
        for ndim in [2, 3]:  # , 4]:
            pts, le, re = make_points(0, ndim)
            self.param_runs += [
                ((pts,), {}),
                ((pts,), {'periodic': True,
                          'left_edge': le, 'right_edge': re}),
                ]
        pts, le, re = make_points(0, 2)
        self.param_raises = [
            (ValueError, (np.zeros((3, 3, 3)),), {}),
            (ValueError, (pts,), {'left_edge': np.zeros(3)}),
            (ValueError, (pts,), {'right_edge': np.zeros(3)}),
            (ValueError, (pts,), {'left_edge': np.zeros((2, 2, 2))}),
            (ValueError, (pts,), {'right_edge': np.zeros((2, 2, 2))}),
            ]



cimport numpy as np
from libcpp.vector cimport vector
from libcpp.set cimport set as cset
from libcpp.pair cimport pair
from libcpp cimport bool as cbool
from libc.stdint cimport uint32_t, uint64_t, int32_t, int64_t, uint8_t


cdef extern from "c_parallel_delaunayD.hpp":
    cdef int VALID

    cdef cppclass ParallelDelaunay_with_info_D[Info] nogil:
        ParallelDelaunay_with_info_D()
        ParallelDelaunay_with_info_D(uint32_t ndim0, double *le0, double *re0,
                                     cbool *periodic0, const char *unique_str0)
        ParallelDelaunay_with_info_D(uint32_t ndim0, double *le0, double *re0,
                                     cbool *periodic0, int limit_mem)
        ParallelDelaunay_with_info_D(uint32_t ndim0, double *le0, double *re0,
                                     cbool *periodic0, int limit_mem,
                                     const char *unique_str0)

        int rank
        int size
        uint32_t ndim
        int limit_mem
        uint64_t npts_total
        uint64_t *idx_total
        Info *info_total
        double *pts_total

        void insert(uint64_t npts, double *pts)

        uint64_t num_cells()
        void consolidate_vols(double *vols)
        uint64_t consolidate_tess(uint64_t tot_ncells_total, Info *tot_idx_inf,
                                  Info *allverts, Info *allneigh)



r"""Plotting routines."""
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from mpl_toolkits.mplot3d.art3d import Line3DCollection


def plot2D(T, plotfile=None, vertex_kw={}, edge_kw={},
           axs=None, subplot_kw={}, gridspec_kw={}, fig_kw={}, save_kw={},
           title=None, xlabel='x', ylabel='y'):
    r"""Plot a 2D triangulation

    Args:
        T (:class:`Delauany2.Delaunay2`): 2D triangulation class.
        plotfile (:obj:`str`, optional): Full path to file where the plot
            should be saved. If None, the plot is displayed. Defaults to None.

        vertex_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.scatter`. Defaults to empty dict.
        edge_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.collections.LineCollection`. Defaults to empty
            dict.

        axs (:obj:`matplotlib.pyplot.Axes`, optional): Axes that should be used
            for plotting. Defaults to None and new axes are created.
        subplot_kw (:obj:`dict`, optional): Keywords passed directly to
            :meth:`matplotlib.figure.Figure.add_subplot`. Defaults to {}.
        gridspec_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.gridspec.GridSpec`. Defaults to empty dict.
        fig_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.figure`. Defaults to empty dict.
        save_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.savefig`. Defaults to empty dict.

        title (:obj:`str`, optional): Title that the plot should be given.
            Defaults to None and no title is displayed.
        xlabel (:obj:`str`, optional): Label for the x-axis. Defaults to 'x'.
        ylabel (:obj:`str`, optional): Label for the y-axis. Defaults to 'y'.

    """
    # Axes creation
    if axs is None:
        plt.close('all')
        fig, axs = plt.subplots(subplot_kw=subplot_kw, gridspec_kw=gridspec_kw,
                                **fig_kw)

    # Labels
    if title is not None:
        axs.set_title(title)
    axs.set_xlabel(xlabel)
    axs.set_ylabel(ylabel)

    # Plot vertices
    v = T.vertices
    axs.scatter(v[:, 0], v[:, 1], **vertex_kw)

    # Plot edges
    e = T.edges
    seg = [v[e[i, :], :] for i in xrange(e.shape[0])]
    lc = LineCollection(seg, **edge_kw)
    axs.add_collection(lc)

    axs.autoscale()
    axs.margins(0.1)

    # Save
    if plotfile is not None:
        plt.savefig(plotfile, **save_kw)
    else:
        plt.show()

    # Return axes
    return axs


def plot3D(T, plotfile=None, vertex_kw={}, edge_kw={},
           axs=None, subplot_kw={}, gridspec_kw={}, fig_kw={}, save_kw={},
           title=None, xlabel='x', ylabel='y', zlabel='z'):
    r"""Plot a 3D triangulation.

    Args:
        T (:class:`Delauany3.Delaunay3`): 3D triangulation class.
        plotfile (:obj:`str`, optional): Full path to file where the plot
            should be saved. If None, the plot is displayed. Defaults to None.

        vertex_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`mpl_toolkits.mplot3d.Axes3D.scatter`. Defaults to {}.
        edge_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`mpl_toolkits.mplot3d.art3d.Line3DCollection`. Defaults to
            empty dict.

        axs (:obj:`matplotlib.pyplot.Axes`, optional): Axes that should be used
            for plotting. Defaults to None and new axes are created.
        subplot_kw (:obj:`dict`, optional): Keywords passed directly to
            :meth:`matplotlib.figure.Figure.add_subplot`. Defaults to {}.
        gridspec_kw (:obj:`dict`, optional): Keywords passed directly to
            :class:`matplotlib.gridspec.GridSpec`. Defaults to empty dict.
        fig_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.figure`. Defaults to empty dict.
        save_kw (:obj:`dict`, optional): Keywords passed directly to
            :func:`matplotlib.pyplot.savefig`. Defaults to empty dict.

        title (:obj:`str`, optional): Title that the plot should be given.
            Defaults to None and no title is displayed.
        xlabel (:obj:`str`, optional): Label for the x-axis. Defaults to 'x'.
        ylabel (:obj:`str`, optional): Label for the y-axis. Defaults to 'y'.
        zlabel (:obj:`str`, optional): Label for the z-axis. Defaults to 'z'.

    """
    # Axes creation
    if axs is None:
        plt.close('all')
        subplot_kw['projection'] = '3d'
        fig, axs = plt.subplots(subplot_kw=subplot_kw, gridspec_kw=gridspec_kw,
                                **fig_kw)

    # Labels
    if title is not None:
        axs.set_title(title)
    axs.set_xlabel(xlabel)
    axs.set_ylabel(ylabel)
    axs.set_zlabel(zlabel)

    # Plot vertices
    v = T.vertices
    axs.scatter(v[:, 0], v[:, 1], v[:, 2], **vertex_kw)

    # Plot edges
    e = T.edges
    seg = [v[e[i, :], :] for i in xrange(e.shape[0])]
    lc = Line3DCollection(seg, **edge_kw)
    axs.add_collection(lc)

    axs.autoscale()
    axs.margins(0.1)

    # Save
    if plotfile is not None:
        plt.savefig(plotfile, **save_kw)
    else:
        plt.show()

    # Return axes
    return axs



r"""Tests for parallel implementation of triangulations."""
import nose.tools as nt
import numpy as np
import os
import time
from cgal4py import parallel, delaunay
from cgal4py.domain_decomp import GenericTree
from test_cgal4py import make_points, make_test, MyTestCase
import multiprocessing as mp
from mpi4py import MPI
import ctypes
np.random.seed(10)


@nt.nottest
def lines_load_test(npts, ndim, periodic=False):
    lines = [
        "from cgal4py.tests.test_cgal4py import make_points",
        "pts, le, re = make_points({}, {})".format(npts, ndim),
        "load_dict = dict(pts=pts, left_edge=le, right_edge=re,",
        "                 periodic={})".format(periodic)]
    return lines


class TestGetMPIType(MyTestCase):

    def setup_param(self):
        self._func = parallel._get_mpi_type
        self.param_equal = [(MPI.INT, ['i'], {}),
                            (MPI.LONG, ['l'], {}),
                            (MPI.FLOAT, ['f'], {}),
                            (MPI.DOUBLE, ['d'], {})]
        self.param_raises = [(ValueError, ['m'], {})]


class TestWriteMPIScript(MyTestCase):

    def setup_param(self):
        self._func = parallel.write_mpi_script
        fname = 'test_mpi_script.py'
        read_lines = lines_load_test(10, 2)
        self.param_runs = [
            ((fname, read_lines, 'triangulate'), {}),
            ((fname, read_lines, 'triangulate'), dict(use_double=True)),
            ((fname, read_lines, 'triangulate'), dict(use_buffer=True)),
            ((fname, read_lines, 'triangulate'), dict(profile=True))]
        self._fname = fname
        self._read_lines = read_lines

    def check_runs(self, args, kwargs):
        self.func(*args, **kwargs)
        assert(os.path.isfile(args[0]))
        os.remove(args[0])

    def test_overwrite(self):
        self.func(self._fname, self._read_lines, 'volumes')
        t0 = os.path.getmtime(self._fname)
        time.sleep(1)
        self.func(self._fname, self._read_lines, 'volumes', overwrite=False)
        t1 = os.path.getmtime(self._fname)
        nt.eq_(t0, t1)
        time.sleep(1)
        self.func(self._fname, self._read_lines, 'volumes', overwrite=True)
        t2 = os.path.getmtime(self._fname)
        nt.assert_not_equal(t1, t2)
        os.remove(self._fname)


class TestParallelLeaf(MyTestCase):

    def setup_param(self):
        self._func = parallel.ParallelLeaf
        self.param_runs = [
            ((0, 2), {}),
            ((0, 3), {}),
            # ((0, 4), {}),
            ((0, 2), {'periodic':True}),
            ((0, 3), {'periodic':True}),
            # ((0, 4), {'periodic':True}),
            ]

    def check_runs(self, args, kwargs):
        pts, tree = make_test(*args, **kwargs)
        left_edges = np.vstack([leaf.left_edge for leaf in tree.leaves])
        right_edges = np.vstack([leaf.right_edge for leaf in tree.leaves])
        for leaf in tree.leaves:
            pleaf = self._func(leaf, left_edges, right_edges)

    def check_tessellate(self, args, kwargs):
        pts, tree = make_test(*args, **kwargs)
        left_edges = np.vstack([leaf.left_edge for leaf in tree.leaves])
        right_edges = np.vstack([leaf.right_edge for leaf in tree.leaves])
        leaf = tree.leaves[0]
        pleaf = self._func(leaf, left_edges, right_edges)
        pleaf.pts = pts[tree.idx[pleaf.idx], :]
        pleaf.tessellate()
        pleaf.tessellate(pts)
        pleaf.tessellate(pts, tree.idx)

    def check_exchange(self, args, kwargs):
        pts, tree = make_test(*args, **kwargs)
        left_edges = np.vstack([leaf.left_edge for leaf in tree.leaves])
        right_edges = np.vstack([leaf.right_edge for leaf in tree.leaves])
        leaf0 = tree.leaves[0]
        leaf1 = tree.leaves[1]
        pleaf0 = self._func(leaf0, left_edges, right_edges)
        pleaf1 = self._func(leaf1, left_edges, right_edges)
        pleaf0.tessellate(pts, tree.idx)
        pleaf1.tessellate(pts, tree.idx)
        out0 = pleaf0.outgoing_points()
        out1 = pleaf1.outgoing_points()
        pleaf1.incoming_points(0, out0[0][1], out0[1], out0[2], out0[3],
                               pts[out0[0][1], :])
        pleaf1.incoming_points(1, out1[0][1], out1[1], out1[2], out1[3],
                               pts[out1[0][1], :])
        pleaf0.incoming_points(0, out0[0][0], out0[1], out0[2], out0[3],
                               pts[out0[0][0], :])
        pleaf0.incoming_points(1, out1[0][0], out1[1], out1[2], out1[3],
                               pts[out1[0][0], :])
        if kwargs.get('periodic', True):
            idx = pleaf1.idx
            pos = pts[idx, :]
            pleaf1.periodic_left[0] = True
            pleaf1.periodic_right[0] = True
            pleaf1.left_neighbors.append(0)
            pos[0,0] = tree.left_edge[0]
            pos[1,0] = tree.right_edge[0]
            pleaf1.incoming_points(1, idx, pleaf1.neighbors, 
                                   pleaf1.left_edges, pleaf1.right_edges, pos)
            pleaf1.incoming_points(0, idx, pleaf0.neighbors, 
                                   pleaf0.left_edges, pleaf0.right_edges, pos)

    def test_tessellate_generator(self):
        for args, kwargs in self.param_runs:
            yield self.check_tessellate, args, kwargs

    def test_exchange_generator(self):
        for args, kwargs in self.param_runs:
            yield self.check_exchange, args, kwargs

    def test_serialize(self):
        pts, tree = make_test(0, 2)
        left_edges = np.vstack([leaf.left_edge for leaf in tree.leaves])
        right_edges = np.vstack([leaf.right_edge for leaf in tree.leaves])
        leaf = tree.leaves[0]
        pleaf = self._func(leaf, left_edges, right_edges)
        pleaf.tessellate(pts, tree.idx)
        pleaf.serialize(store=True)


class TestDelaunayProcessMPI(MyTestCase):

    def setup_param(self):
        self._func = parallel.DelaunayProcessMPI
        taskname1 = 'triangulate'
        taskname2 = 'volumes'
        ndim = 2
        periodic = False
        pts, tree = make_test(0, ndim, periodic=periodic)
        le = tree.left_edge
        re = tree.right_edge
        self._pts = pts
        self._tree = tree
        # self._dummy3 = self._func(taskname2, pts, tree)
        # self._dummy4 = self._func(taskname2, pts, tree, use_buffer=True)
        # self._leaves = self._dummy1._leaves
        # TODO: use_buffer curregntly segfaults when run with coverage
        self.param_runs = [
            # Using C++ communications
            # ((taskname1, pts), {}),
            # ((taskname1, pts, left_edge=le, right_edge=re), {}),
            # ((taskname1, pts), {'use_double':True}),
            # ((taskname1, pts), {'limit_mem':True}),
            # ((taskname2, pts), {}),
            # ((taskname2, pts), {'limit_mem':True}),
            # Using Python communications
            ((taskname1, pts), {'use_python':True}),
            ((taskname1, pts, tree), {'use_python':True}),
            ((taskname1, pts, GenericTree.from_tree(tree)),
                 {'use_python':True}),
            ((taskname1, pts, tree), {'use_python':True,
                                      'use_double':True}),
            # ((taskname1, pts, tree), {'use_python':True},
            #                           'use_buffer':True}),
            ((taskname1, pts, tree), {'use_python':True,
                                      'limit_mem':True}),
            ((taskname2, pts, tree), {'use_python':True}),
            # ((taskname2, pts, tree), {'use_python':True},
            #                           'use_buffer':True}),
            ((taskname2, pts, tree), {'use_python':True,
                                      'limit_mem':True}),
            ]
        self.param_raises = [
            (ValueError, ('null', pts, tree), {})
            ]

    def check_runs(self, args, kwargs):
        x = self.func(*args, **kwargs)
        x.run()
        fname = x.output_filename()
        if os.path.isfile(fname):
            os.remove(fname)

    def test_gather_leaf_arrays(self):
        taskname1 = 'triangulate'
        pts = self._pts
        tree = self._tree
        dummy1 = self.func(taskname1, pts, tree, use_python=True)
        dummy2 = self.func(taskname1, pts, tree, use_python=True,
                           use_buffer=True)
        leaves = tree.leaves
        arr = {leaf.id: np.arange(5*(leaf.id+1)) for leaf in leaves}
        dummy1.gather_leaf_arrays(arr)
        dummy2.gather_leaf_arrays(arr)


class TestDelaunayProcessMulti(MyTestCase):

    def setup_param(self):
        self._func = parallel.DelaunayProcessMulti
        self.param_runs = [
            (('triangulate',), {}),
            (('triangulate',), {'limit_mem':True}),
            (('volumes',), {}),
            (('triangulate',), {'limit_mem':True}),
            ]

    def check_runs(self, args, kwargs):
        (taskname,) = args
        ndim = 2
        periodic = False
        pts, tree = make_test(0, ndim, periodic=periodic)
        idxArray = mp.RawArray(ctypes.c_ulonglong, tree.idx.size)
        ptsArray = mp.RawArray('d', pts.size)
        memoryview(idxArray)[:] = tree.idx
        memoryview(ptsArray)[:] = pts
        left_edges = np.vstack([leaf.left_edge for leaf in tree.leaves])
        right_edges = np.vstack([leaf.right_edge for leaf in tree.leaves])
        leaves = tree.leaves
        nproc = 2  # len(leaves)
        count = [mp.Value('i',0),mp.Value('i',0),mp.Value('i',0)]
        lock = mp.Condition()
        queues = [mp.Queue() for _ in xrange(nproc+1)]
        in_pipes = [None for _ in xrange(nproc)]
        out_pipes = [None for _ in xrange(nproc)]
        for i in range(nproc):
            out_pipes[i],in_pipes[i] = mp.Pipe(True)
        # Split leaves
        task2leaves = [[] for _ in xrange(nproc)]
        for leaf in leaves:
            task = leaf.id % nproc
            task2leaves[task].append(leaf)
        # Dummy process
        processes = []
        for i in xrange(nproc):
            processes.append(self._func(
                taskname, i, task2leaves[i],
                ptsArray, idxArray, left_edges, right_edges,
                queues, lock, count, in_pipes[i], **kwargs))
        # Perform setup on higher processes
        for i in xrange(1, nproc):
            P = processes[i]
            P.tessellate_leaves()
            P.outgoing_points()
        for i in xrange(1, nproc):
            count[0].value += 1
        # Perform entire run on lowest process
        P = processes[0]
        P.run(test_in_serial=True)
        # Do tear down on higher processes
        for i in xrange(1, nproc):
            P = processes[i]
            P.incoming_points()
            P.enqueue_result()
            for l in range(len(task2leaves[i])):
                x = P.receive_result(out_pipes[i])
        # Clean up files
        for i in xrange(nproc):
            P = processes[i]
            for leaf in P._leaves:
                if kwargs.get('limit_mem', False):
                    leaf.remove_tess()
                ffinal = leaf.tess_output_filename
                if os.path.isfile(ffinal):
                    os.remove(ffinal)


class TestParallelDelaunay(MyTestCase):

    def setup_param(self):
        self._func = parallel.ParallelDelaunay
        ndim_list = [2, 3]  # , 4]
        param_test = []
        self._fprof = 'test_ParallelDelaunay.cProfile'
        for ndim in ndim_list:
            param_test += [
                ((0, ndim, 2), {}),
                ((100, ndim, 2), {'nleaves': 2}),
                ((100, ndim, 4), {'nleaves': 4}),
                ((100, ndim, 4), {'nleaves': 8}),
                # ((1000, ndim, 2), {'nleaves': 2}),
                # ((4*4*2, ndim, 4), {'leafsize': 8}),
                # ((1e5, ndim, 8), {'nleaves': 8}),
                # ((1e7, ndim, 10), {'nleaves': 10}),
                ]
        self.param_returns = []
        for args, kwargs in param_test:
            pts, tree = make_test(args[0], args[1], **kwargs)
            ans = delaunay.Delaunay(pts)
            read_lines = lines_load_test(args[0], args[1])
            for limit_mem in [False, True]:
                self.param_returns += [
                    (ans, (pts, tree, args[2]),
                         {'use_mpi': False, 'limit_mem': limit_mem})
                    ]
                for profile in [False, self._fprof]:
                    self.param_returns += [
                        (ans, (pts, tree, args[2]),
                             {'use_mpi': True, 'limit_mem': limit_mem,
                              'profile': profile, 'use_python':True,
                              'use_buffer': False}),
                        (ans, (pts, tree, args[2]),
                             {'use_mpi': True, 'limit_mem': limit_mem,
                              'profile': profile, 'use_python':True,
                              'use_buffer': True}),
                        # (ans, (pts, tree, args[2]),
                        #      {'use_mpi': True, 'limit_mem': limit_mem,
                        #       'profile': profile, 'use_python':False})
                        ]

    def check_returns(self, result, args, kwargs):
        T_seri = result
        T_para = self.func(*args, **kwargs)
        ndim = args[0].shape[1]
        c_seri, n_seri, inf_seri = T_seri.serialize(sort=True)
        c_para, n_para, inf_para = T_para.serialize(sort=True)
        try:
            assert(np.all(c_seri == c_para))
            assert(np.all(n_seri == n_para))
            assert(T_para.is_equivalent(T_seri))
        except:
            for name, T in zip(['Parallel','Serial'],[T_para, T_seri]):
                print(name)
                print('\t verts: {}, {}, {}'.format(
                    T.num_verts, T.num_finite_verts, T.num_infinite_verts))
                print('\t cells: {}, {}, {}'.format(
                    T.num_cells, T.num_finite_cells, T.num_infinite_cells))
                print('\t edges: {}, {}, {}'.format(
                    T.num_edges, T.num_finite_edges, T.num_infinite_edges))
                if ndim == 3:
                    print('\t facets: {}, {}, {}'.format(
                        T.num_facets, T.num_finite_facets,
                        T.num_infinite_facets))
            raise
        if os.path.isfile(self._fprof):
            os.remove(self._fprof)


class TestParallelVoronoiVolumes(MyTestCase):

    def setup_param(self):
        self._func = parallel.ParallelVoronoiVolumes
        ndim_list = [2, 3]  # , 4]
        param_test = []
        self._fprof = 'test_ParallelVoronoiVolumes.cProfile'
        for ndim in ndim_list:
            param_test += [
                ((0, ndim, 2), {}),
                ((100, ndim, 2), {'nleaves': 2}),
                ((100, ndim, 4), {'nleaves': 4}),
                ((100, ndim, 4), {'nleaves': 8}),
                # ((1000, ndim, 2), {'nleaves': 2}),
                # ((4*4*2, ndim, 4), {'leafsize': 8}),
                # ((1e5, ndim, 8), {'nleaves': 8}),
                # ((1e7, ndim, 10), {'nleaves': 10}),
                ]
        self.param_returns = []
        for args, kwargs in param_test:
            pts, tree = make_test(args[0], args[1], **kwargs)
            ans = delaunay.VoronoiVolumes(pts)
            read_lines = lines_load_test(args[0], args[1])
            for limit_mem in [False, True]:
                self.param_returns += [
                    (ans, (pts, tree, args[2]),
                         {'use_mpi': False, 'limit_mem': limit_mem})
                    ]
                for profile in [False, self._fprof]:
                    self.param_returns += [
                        (ans, (pts, tree, args[2]),
                             {'use_mpi': True, 'limit_mem': limit_mem,
                              'profile': profile, 'use_python':True,
                              'use_buffer': False}),
                        (ans, (pts, tree, args[2]),
                             {'use_mpi': True, 'limit_mem': limit_mem,
                              'profile': profile, 'use_python':True,
                              'use_buffer': True}),
                        # (ans, (pts, tree, args[2]),
                        #      {'use_mpi': True, 'limit_mem': limit_mem,
                        #       'profile': profile, 'use_python':False})
                        ]

    def check_returns(self, result, args, kwargs):
        print result
        print self.func(*args, **kwargs)
        assert(np.allclose(result, self.func(*args, **kwargs)))
        if os.path.isfile(self._fprof):
            os.remove(self._fprof)



import numpy as np
cimport numpy as np
cimport cython
from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libc.stdint cimport uint32_t, uint64_t, int64_t, int32_t
from libcpp cimport bool as cbool
from cpython cimport bool as pybool
import copy

import scipy
from scipy.sparse import csr_matrix

def py_max_pts(np.ndarray[np.float64_t, ndim=2] pos):
    r"""Get the maximum of points along each coordinate. 

    Args: 
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 

    Returns: 
        np.ndarray of float64: Maximum of pos along each coordinate. 

    """
    cdef uint64_t n = <uint64_t>pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef np.float64_t* cout = max_pts(&pos[0,0], n, m)
    cdef uint32_t i = 0
    cdef np.ndarray[np.float64_t] out = np.zeros(m, 'float64')
    for i in range(m):
        out[i] = cout[i]
    return out

def py_min_pts(np.ndarray[np.float64_t, ndim=2] pos):
    r"""Get the minimum of points along each coordinate. 

    Args: 
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 

    Returns: 
        np.ndarray of float64: Minimum of pos along each coordinate. 

    """
    cdef uint64_t n = <uint64_t>pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef np.float64_t* cout = min_pts(&pos[0,0], n, m)
    cdef uint32_t i = 0
    cdef np.ndarray[np.float64_t] out = np.zeros(m, 'float64')
    for i in range(m):
        out[i] = cout[i]
    return out

def py_quickSort(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the indices required to sort coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 
        d (np.uint32_t): Dimension that pos should be sorted along.

    Returns:
        np.ndarray of uint64: Indices that sort pos along dimension d.

    """
    cdef uint32_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef np.ndarray[np.uint64_t, ndim=1] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    quickSort(&pos[0,0], &idx[0], ndim, d, l, r)
    return idx

def py_insertSort(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the indices required to sort coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 
        d (np.uint32_t): Dimension that pos should be sorted along.

    Returns:
        np.ndarray of uint64: Indices that sort pos along dimension d.

    """
    cdef uint32_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef np.ndarray[np.uint64_t, ndim=1] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    insertSort(&pos[0,0], &idx[0], ndim, d, l, r)
    return idx

def py_pivot(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the index of the median of medians along one dimension and indices 
    that partition pos according to the median of medians.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 
        d (np.uint32_t): Dimension that pos should be partitioned along.

    Returns:
        tuple of int64 and np.ndarray of uint64: Index q of idx that is the 
            median of medians & array of indices that partition pos such that 
            the first q elements are less than the median of medians.

    """
    cdef uint32_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef np.ndarray[np.uint64_t, ndim=1] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef int64_t q = pivot(&pos[0,0], &idx[0], ndim, d, l, r)
    return q+1, idx

def py_partition(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d,
                 np.int64_t p):
    r"""Get the indices required to partition coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates. 
        d (np.uint32_t): Dimension that pos should be partitioned along.
        p (np.int64_t): Element of pos[:,d] that pos should be used to partition 
            pos.

    Returns:
        tuple of int64 and np.ndarray of uint64: Number of elements q in pos 
            that are less that or equal to the pivot & array of indices that 
            partition pos such that the first q elements are less than the 
            pivot and the last (n-q) elements are greater than the pivot.

    """
    cdef uint32_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef np.ndarray[np.uint64_t, ndim=1] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef int64_t q = partition(&pos[0,0], &idx[0], ndim, d, l, r, p)
    return q+1, idx

def py_select(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d,
              np.int64_t t):
    r"""Get the indices required to partition coordiantes such that the first 
    t elements in pos[:,d] are the smallest t elements in pos[:,d].

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be partitioned along. 
        t (np.int64_t): Number of smallest elements in pos[:,d] that should be 
            partitioned.

    Returns:
        np.ndarray of uint64: Indices required to partition pos such that the 
            1st t elements in pos[:,d] are the smallest t elements in pos[:,d].

    """
    cdef uint32_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef np.ndarray[np.uint64_t, ndim=1] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef int64_t q = select(&pos[0,0], &idx[0], ndim, d, l, r, t)
    return idx


